#
#
#
from .record import YandexCloudAnameRecord
from .version import __VERSION__, __version__
from .yandex360_provider import Yandex360Provider
from .yandexcloud_cdn_source import YandexCloudCDNSource
from .yandexcloud_cm_source import YandexCloudCMSource
from .yandexcloud_provider import YandexCloudProvider

__all__ = [
    'YandexCloudProvider',
    'Yandex360Provider',
    'YandexCloudAnameRecord',
    'YandexCloudCMSource',
    'YandexCloudCDNSource',
]

# quell warnings
__VERSION__
__version__



from google.protobuf.timestamp_pb2 import Timestamp
from yandex.cloud.certificatemanager.v1.certificate_pb2 import (
    Certificate,
    CertificateType,
    Challenge,
    ChallengeType,
)

from tests.fixtures import STUB_FOLDER_ID

STUB_CERTIFICATE_1 = Certificate(
    id="fpqahphei3aegootoh0a",
    folder_id=STUB_FOLDER_ID,
    created_at=Timestamp(seconds=1711559410, nanos=544000000),
    name="wildcard-cert",
    description="Wildcard cert for domain",
    type=CertificateType.MANAGED,
    domains=["*.example.com", "example.com"],
    status=Certificate.Status.ISSUED,
    issuer="CN=R3,O=Let\'s Encrypt,C=US",
    subject="CN=*.example.com",
    serial="9492fe88f263d58e0b686885e8c98c0eaee",
    updated_at=Timestamp(seconds=1711559421, nanos=701000000),
    issued_at=Timestamp(seconds=1711559428, nanos=516000000),
    not_after=Timestamp(seconds=1716559440),
    not_before=Timestamp(seconds=1711559455),
    challenges=[
        Challenge(
            domain="example.com",
            type=ChallengeType.DNS,
            created_at=Timestamp(seconds=1711559410, nanos=487000000),
            updated_at=Timestamp(seconds=1711559428, nanos=789000000),
            status=Challenge.Status.VALID,
            message="Create a record in your DNS provider.",
            dns_challenge=Challenge.DnsRecord(
                name="_acme-challenge.example.com.",
                type="CNAME",
                value="fpqahphei3aegootoh0a.cm.yandexcloud.net.",
            ),
        ),
        Challenge(
            domain="example.com",
            type=ChallengeType.DNS,
            created_at=Timestamp(seconds=1711559410, nanos=487000000),
            updated_at=Timestamp(seconds=1711559428, nanos=789000000),
            status=Challenge.Status.VALID,
            message="Create a record in your DNS provider.",
            dns_challenge=Challenge.DnsRecord(
                name="_acme-challenge.example.com.",
                type="TXT",
                value="Hw6x5Kj48Xu64sNMxHd7zBrEvetzQsJAXR8wFI8zytC",
            ),
        ),
        Challenge(
            domain="example.com",
            type=ChallengeType.DNS,
            created_at=Timestamp(seconds=1711559410, nanos=487000000),
            updated_at=Timestamp(seconds=1711559428, nanos=789000000),
            status=Challenge.Status.VALID,
            message="Create a record in your DNS provider.",
            dns_challenge=Challenge.DnsRecord(
                name="_acme-challenge.example.com.",
                type="CNAME",
                value="fpqahphei3aegootoh0a.cm.yandexcloud.net.",
            ),
        ),
        Challenge(
            domain="example.com",
            type=ChallengeType.DNS,
            created_at=Timestamp(seconds=1711559410, nanos=487000000),
            updated_at=Timestamp(seconds=1711559428, nanos=789000000),
            status=Challenge.Status.VALID,
            message="Create a record in your DNS provider.",
            dns_challenge=Challenge.DnsRecord(
                name="_acme-challenge.example.com.",
                type="TXT",
                value="EwOftdMfheciHoB1L3M1cKz2NbjyFlZEMEPZ6d9N8rP",
            ),
        ),
    ],
    deletion_protection=True,
)

STUB_CERTIFICATE_2 = Certificate(
    id="fpqco0ongail3feejee1",
    folder_id="b1g0aj2nmme3adupg65e",
    created_at=Timestamp(seconds=1711559410, nanos=544000000),
    name="double-cdn",
    description="Double CDN",
    type=CertificateType.MANAGED,
    domains=["cdn.example.com", "cdn2.example.com"],
    status=Certificate.Status.ISSUED,
    issuer="CN=R3,O=Let\'s Encrypt,C=US",
    subject="CN=cdn.example.com",
    serial="40bc979fc51c312c66f13621c9c53e84d51",
    updated_at=Timestamp(seconds=1711559421, nanos=701000000),
    issued_at=Timestamp(seconds=1711559428, nanos=516000000),
    not_after=Timestamp(seconds=1716559440),
    not_before=Timestamp(seconds=1711559455),
    challenges=[
        Challenge(
            domain="cdn.example.com",
            type=ChallengeType.DNS,
            created_at=Timestamp(seconds=1711559410, nanos=487000000),
            updated_at=Timestamp(seconds=1711559428, nanos=789000000),
            status=Challenge.Status.VALID,
            message="Create a record in your DNS provider.",
            dns_challenge=Challenge.DnsRecord(
                name="_acme-challenge.cdn.example.com.",
                type="CNAME",
                value="fpqco0ongail3feejee1.cm.yandexcloud.net.",
            ),
        ),
        Challenge(
            domain="cdn.example.com",
            type=ChallengeType.DNS,
            created_at=Timestamp(seconds=1711559410, nanos=487000000),
            updated_at=Timestamp(seconds=1711559428, nanos=789000000),
            status=Challenge.Status.VALID,
            message="Create a record in your DNS provider.",
            dns_challenge=Challenge.DnsRecord(
                name="_acme-challenge.cdn.example.com.",
                type="TXT",
                value="DIzWSC8sf8Sh2iMMO7chYopq8aAiNhcuxINGDKvONEA",
            ),
        ),
        Challenge(
            domain="cdn2.example.com",
            type=ChallengeType.DNS,
            created_at=Timestamp(seconds=1711559410, nanos=487000000),
            updated_at=Timestamp(seconds=1711559428, nanos=789000000),
            status=Challenge.Status.VALID,
            message="Create a record in your DNS provider.",
            dns_challenge=Challenge.DnsRecord(
                name="_acme-challenge.cdn2.example.com.",
                type="CNAME",
                value="fpqco0ongail3feejee1.cm.yandexcloud.net.",
            ),
        ),
        Challenge(
            domain="cdn2.example.com",
            type=ChallengeType.DNS,
            created_at=Timestamp(seconds=1711559410, nanos=487000000),
            updated_at=Timestamp(seconds=1711559428, nanos=789000000),
            status=Challenge.Status.VALID,
            message="Create a record in your DNS provider.",
            dns_challenge=Challenge.DnsRecord(
                name="_acme-challenge.cdn2.example.com.",
                type="TXT",
                value="8RaeDrdPzsG6L0atqh5w1qcq3DHudQ7EZCgyQfTlA6i",
            ),
        ),
    ],
    deletion_protection=True,
)

STUB_IDNA_CERTIFICATE = Certificate(
    id="fpqix1ahwo3ijiebuu0i",
    folder_id="b1g0aj2nmme3adupg65e",
    created_at=Timestamp(seconds=1711559410, nanos=544000000),
    name="idna-test",
    description="Another idna test",
    type=CertificateType.MANAGED,
    domains=["*.xn--e1aybc.xn--p1ai", "xn--e1aybc.xn--p1ai"],
    status=Certificate.Status.ISSUED,
    issuer="CN=R3,O=Let\'s Encrypt,C=US",
    subject="CN=*.xn--e1aybc.xn--p1ai",
    serial="1929d9f7c81f25c32830ebfe29fec2b247a",
    updated_at=Timestamp(seconds=1711559421, nanos=701000000),
    issued_at=Timestamp(seconds=1711559428, nanos=516000000),
    not_after=Timestamp(seconds=1716559440),
    not_before=Timestamp(seconds=1711559455),
    challenges=[
        Challenge(
            domain="xn--e1aybc.xn--p1ai",
            type=ChallengeType.DNS,
            created_at=Timestamp(seconds=1711559410, nanos=487000000),
            updated_at=Timestamp(seconds=1711559428, nanos=789000000),
            status=Challenge.Status.VALID,
            message="Create a record in your DNS provider.",
            dns_challenge=Challenge.DnsRecord(
                name="_acme-challenge.xn--e1aybc.xn--p1ai.",
                type="CNAME",
                value="fpqix1ahwo3ijiebuu0i.cm.yandexcloud.net.",
            ),
        ),
        Challenge(
            domain="xn--e1aybc.xn--p1ai",
            type=ChallengeType.DNS,
            created_at=Timestamp(seconds=1711559410, nanos=487000000),
            updated_at=Timestamp(seconds=1711559428, nanos=789000000),
            status=Challenge.Status.VALID,
            message="Create a record in your DNS provider.",
            dns_challenge=Challenge.DnsRecord(
                name="_acme-challenge.xn--e1aybc.xn--p1ai.",
                type="TXT",
                value="r4joRe9NgOIhyYLbNOCxdSPXNjnuU95kD1VbYbM0sNd",
            ),
        ),
        Challenge(
            domain="xn--e1aybc.xn--p1ai",
            type=ChallengeType.DNS,
            created_at=Timestamp(seconds=1711559410, nanos=487000000),
            updated_at=Timestamp(seconds=1711559428, nanos=789000000),
            status=Challenge.Status.VALID,
            message="Create a record in your DNS provider.",
            dns_challenge=Challenge.DnsRecord(
                name="_acme-challenge.xn--e1aybc.xn--p1ai.",
                type="CNAME",
                value="fpqix1ahwo3ijiebuu0i.cm.yandexcloud.net.",
            ),
        ),
        Challenge(
            domain="xn--e1aybc.xn--p1ai",
            type=ChallengeType.DNS,
            created_at=Timestamp(seconds=1711559410, nanos=487000000),
            updated_at=Timestamp(seconds=1711559428, nanos=789000000),
            status=Challenge.Status.VALID,
            message="Create a record in your DNS provider.",
            dns_challenge=Challenge.DnsRecord(
                name="_acme-challenge.xn--e1aybc.xn--p1ai.",
                type="TXT",
                value="ShsQrb08EUXyAmJ2UVGs5LPZuoOJEwTTmhQRM5LRq9T",
            ),
        ),
    ],
    deletion_protection=True,
)

IMPORTED_CERTIFICATE = Certificate(
    id="fpqeiqu3yoose7ohnuuf",
    folder_id=STUB_FOLDER_ID,
    created_at=Timestamp(seconds=1711639027, nanos=815000000),
    name="test-cert",
    type=CertificateType.IMPORTED,
    domains="example",
    status=Certificate.Status.ISSUED,
    issuer="CN=example",
    subject="CN=example",
    serial="46c86b9e43fad217487cf0883e811af50d61d979",
    updated_at=Timestamp(seconds=1711639027, nanos=815000000),
    issued_at=Timestamp(seconds=1711639027, nanos=815000000),
    not_after=Timestamp(seconds=1873526439),
    not_before=Timestamp(seconds=1558166439),
    incomplete_chain=True,
)

HTTP_CHALLENGE_CERTIFICATE = Certificate(
    id="fpqhahxiowi7kohzeegh",
    folder_id=STUB_FOLDER_ID,
    created_at=Timestamp(seconds=1711640375, nanos=70000000),
    name="test-http",
    type=CertificateType.MANAGED,
    domains=["http.example.com"],
    status=Certificate.Status.VALIDATING,
    updated_at=Timestamp(seconds=1711640375, nanos=70000000),
    challenges=[
        Challenge(
            domain="http.example.com",
            type=ChallengeType.HTTP,
            created_at=Timestamp(seconds=1711640375, nanos=70000000),
            updated_at=Timestamp(seconds=1711640377, nanos=724000000),
            status=Challenge.Status.PENDING,
            message="Create a file in your web server\'s base directory.",
            http_challenge=Challenge.HttpFile(
                url="http://http.example.com/.well-known/acme-challenge/Ocigx_DuIcj50N9tRfUx8sdJC8tCBQ3vRrtUP_rro8ZY",
                content="Ocigx_DuIcj50N9tRfUx8sdJC8tCBQ3vRrtUP_rro8ZY.QRuoX6yZYcbmuw_Gb0DQbwXBFJWZkMr_DMCa2yaVeQ",
            ),
        )
    ],
)



from octodns.record import Record, ValueMixin
from octodns.record.target import _TargetValue


class YandexCloudAnameValue(_TargetValue):
    pass


# Works similar to ALIAS but allows to specify subdomains
class YandexCloudAnameRecord(ValueMixin, Record):
    _type = 'YandexCloudProvider/ANAME'
    _value_type = YandexCloudAnameValue


Record.register_type(YandexCloudAnameRecord)



#
#
#

import grpc
import pytest
import yandexcloud
from yandex.cloud.dns.v1.dns_zone_pb2 import RecordSet
from yandex.cloud.dns.v1.dns_zone_service_pb2 import (
    ListDnsZoneRecordSetsResponse,
    ListDnsZonesResponse,
)
from yandex.cloud.dns.v1.dns_zone_service_pb2_grpc import DnsZoneServiceStub

from octodns.idna import idna_decode
from octodns.provider.plan import Plan
from octodns.record import Create, Delete, Update
from octodns.zone import Zone

from octodns_yandex import YandexCloudProvider
from octodns_yandex.auth import AUTH_TYPE_METADATA
from octodns_yandex.yandexcloud_provider import (
    YandexCloudException,
    map_octodns_to_rset,
    map_rset_to_octodns,
)
from tests.fixtures.dns_zones import (
    STUB_FOLDER_ID,
    STUB_IDNA_ZONE,
    STUB_IDNA_ZONE_NAME,
    STUB_ZONE_1,
    STUB_ZONE_2,
    STUB_ZONE_NAME,
    STUB_ZONE_PUBLIC,
)
from tests.fixtures.record_sets import STUB_RECORDS


class StubSDK:
    def __init__(self, *args, **kwargs):
        pass

    def client(self, *args):
        class StubChannel:
            def unary_unary(self, *args, **kwargs):
                pass

        return DnsZoneServiceStub(StubChannel())

    def wait_operation_and_get_result(self, *args, **kwargs):
        pass


@pytest.fixture()
def disable_sdk(monkeypatch):
    monkeypatch.setattr(yandexcloud, 'SDK', StubSDK)
    return None


@pytest.fixture()
def provider(disable_sdk):
    return YandexCloudProvider(
        "test", folder_id=STUB_FOLDER_ID, auth_type=AUTH_TYPE_METADATA
    )


@pytest.fixture()
def provider_with_zone(monkeypatch, provider):
    def _list(request):
        return ListDnsZonesResponse(
            next_page_token='', dns_zones=[STUB_ZONE_PUBLIC]
        )

    monkeypatch.setattr(provider.dns_service, 'List', _list)
    return provider


class TestYandexCloudProvider:
    class TestFindZone:
        def test_find_zone_found_one(self, monkeypatch, provider):
            def _list(request):
                return ListDnsZonesResponse(
                    next_page_token='', dns_zones=[STUB_IDNA_ZONE]
                )

            monkeypatch.setattr(provider.dns_service, 'List', _list)
            zone_id = provider.get_zone_id_by_name(STUB_IDNA_ZONE_NAME)
            assert zone_id == STUB_IDNA_ZONE.id

        def test_find_zone_mapped(self, monkeypatch, disable_sdk):
            def _list(request):
                return ListDnsZonesResponse(
                    next_page_token='',
                    dns_zones=[STUB_ZONE_1, STUB_ZONE_2, STUB_ZONE_PUBLIC],
                )

            provider = YandexCloudProvider(
                "test",
                folder_id=STUB_FOLDER_ID,
                auth_type=AUTH_TYPE_METADATA,
                zone_ids_map={
                    STUB_ZONE_NAME: STUB_ZONE_2.id,
                    idna_decode(STUB_IDNA_ZONE_NAME): STUB_IDNA_ZONE.id,
                },
            )
            monkeypatch.setattr(provider.dns_service, 'List', _list)
            zone_id = provider.get_zone_id_by_name(STUB_ZONE_NAME)
            assert zone_id == STUB_ZONE_2.id

            def _list_idna(request):
                return ListDnsZonesResponse(
                    next_page_token='',
                    dns_zones=[STUB_ZONE_1, STUB_ZONE_2, STUB_IDNA_ZONE],
                )

            monkeypatch.setattr(provider.dns_service, 'List', _list_idna)
            zone_id = provider.get_zone_id_by_name(STUB_IDNA_ZONE_NAME)
            assert zone_id == STUB_IDNA_ZONE.id

        def test_find_zone_not_found(self, monkeypatch, provider):
            def _list(request):
                return ListDnsZonesResponse(next_page_token='', dns_zones=[])

            monkeypatch.setattr(provider.dns_service, 'List', _list)
            zone_id = provider.get_zone_id_by_name(STUB_ZONE_NAME)
            assert zone_id is None

        def test_find_zone_multiple_public(self, monkeypatch, disable_sdk):
            def _list(request):
                return ListDnsZonesResponse(
                    next_page_token='',
                    dns_zones=[STUB_ZONE_1, STUB_ZONE_PUBLIC, STUB_ZONE_2],
                )

            provider = YandexCloudProvider(
                "test",
                folder_id=STUB_FOLDER_ID,
                auth_type=AUTH_TYPE_METADATA,
                prioritize_public=True,
            )
            monkeypatch.setattr(provider.dns_service, 'List', _list)
            zone_id = provider.get_zone_id_by_name(STUB_ZONE_NAME)
            assert zone_id == STUB_ZONE_PUBLIC.id

            def _list_without_public(request):
                return ListDnsZonesResponse(
                    next_page_token='', dns_zones=[STUB_ZONE_1, STUB_ZONE_2]
                )

            monkeypatch.setattr(
                provider.dns_service, 'List', _list_without_public
            )
            zone_id = provider.get_zone_id_by_name(STUB_ZONE_NAME)
            assert zone_id == STUB_ZONE_1.id

        def test_find_zone_multiple_private(self, monkeypatch, disable_sdk):
            def _list(request):
                return ListDnsZonesResponse(
                    next_page_token='',
                    dns_zones=[STUB_ZONE_PUBLIC, STUB_ZONE_1, STUB_ZONE_2],
                )

            provider = YandexCloudProvider(
                "test",
                folder_id=STUB_FOLDER_ID,
                auth_type=AUTH_TYPE_METADATA,
                prioritize_public=False,
            )
            monkeypatch.setattr(provider.dns_service, 'List', _list)
            zone_id = provider.get_zone_id_by_name(STUB_ZONE_NAME)
            assert zone_id == STUB_ZONE_1.id

    class TestMapping:
        def test_mapping_supported(self):
            zone = Zone(STUB_ZONE_NAME, [])
            for type in YandexCloudProvider.SUPPORTS:
                rset = STUB_RECORDS[type.split('/')[-1]]
                # Make rset copy because of ANAME handling
                record = map_rset_to_octodns(
                    None, zone, False, rset.__deepcopy__()
                )

                recovered_rset = map_octodns_to_rset(record)
                assert rset.name == recovered_rset.name
                assert rset.type == recovered_rset.type
                assert rset.ttl == recovered_rset.ttl
                assert sorted(rset.data) == sorted(recovered_rset.data)

        def test_mapping_lenient(self):
            zone = Zone(STUB_ZONE_NAME, [])
            records = map(
                lambda x: STUB_RECORDS[x],
                (
                    'CAA_quoted',
                    'TXT_quoted',
                    'CNAME_nodot',
                    'CNAME_dot',
                    'MX_nodot',
                    'SRV_nodot',
                    'ANAME_nodot',
                ),
            )
            for rset in records:
                # Just check that it maps
                map_rset_to_octodns(None, zone, True, rset)

        def test_mapping_unknown(self):
            with pytest.raises(YandexCloudException):
                map_rset_to_octodns(
                    None,
                    None,
                    True,
                    RecordSet(
                        name="test.example.com.",
                        type="NON_EXISTENT_TYPE",
                        ttl=300,
                        data=["127.0.0.1", "127.0.0.2"],
                    ),
                )

    class TestPopulate:
        def test_populate_not_found(self, monkeypatch, provider):
            def _list(request):
                return ListDnsZonesResponse(next_page_token='', dns_zones=[])

            monkeypatch.setattr(provider.dns_service, 'List', _list)

            zone = Zone(STUB_ZONE_NAME, [])
            b = provider.populate(zone)
            assert not b
            assert len(zone.records) == 0

        def test_populate_single_page(self, monkeypatch, provider_with_zone):
            def _list_recordsets(request):
                return ListDnsZoneRecordSetsResponse(
                    next_page_token='',
                    record_sets=[STUB_RECORDS['A'], STUB_RECORDS['SOA']],
                )

            monkeypatch.setattr(
                provider_with_zone.dns_service,
                'ListRecordSets',
                _list_recordsets,
            )

            zone = Zone(STUB_ZONE_NAME, [])
            b = provider_with_zone.populate(zone)
            assert b
            assert (
                len(zone.records) == 1
            )  # SOA ignored because it is not on provider's (and octodns) SUPPORTS list

        def test_populate_multiple_pages(self, monkeypatch, provider_with_zone):
            TEST_RECORDS = list(STUB_RECORDS.values())

            def _list_recordsets(request):
                offset = int(request.page_token) if request.page_token else 0
                next_offset = offset + 10
                rsets = TEST_RECORDS[offset:next_offset]
                if next_offset >= len(TEST_RECORDS):
                    next_offset = ''
                return ListDnsZoneRecordSetsResponse(
                    next_page_token=str(next_offset), record_sets=rsets
                )

            monkeypatch.setattr(
                provider_with_zone.dns_service,
                'ListRecordSets',
                _list_recordsets,
            )

            zone = Zone(STUB_ZONE_NAME, [])
            b = provider_with_zone.populate(zone, lenient=True)
            assert b
            assert len(zone.records) == len(STUB_RECORDS) - 1  # without SOA

    class TestApply:
        @staticmethod
        def _make_plan(create, delete, update):
            zone = Zone(STUB_ZONE_NAME, [])
            return Plan(
                existing=None,
                desired=zone,
                changes=[
                    *map(
                        lambda x: Create(
                            map_rset_to_octodns(None, zone, True, x)
                        ),
                        create,
                    ),
                    *map(
                        lambda x: Delete(
                            map_rset_to_octodns(None, zone, True, x)
                        ),
                        delete,
                    ),
                    *map(
                        lambda x: Update(
                            map_rset_to_octodns(None, zone, True, x[0]),
                            map_rset_to_octodns(None, zone, True, x[1]),
                        ),
                        update,
                    ),
                ],
                exists=False,
            )

        @staticmethod
        def _compare_rset_batches(result, expected):
            result_str = sorted(map(lambda x: x.SerializeToString(), result))
            expected_str = sorted(
                map(lambda x: x.SerializeToString(), expected)
            )
            assert result_str == expected_str

        def test_apply_not_found(self, monkeypatch, provider):
            def _list(request):
                return ListDnsZonesResponse(next_page_token='', dns_zones=[])

            monkeypatch.setattr(provider.dns_service, 'List', _list)

            with pytest.raises(YandexCloudException):
                provider.apply(self._make_plan([], [], []))

        def test_apply_simple(self, monkeypatch, provider_with_zone):
            additions, deletions = [], []

            def _update_record_sets(request):
                additions.append(list(request.additions))
                deletions.append(list(request.deletions))
                return None

            monkeypatch.setattr(
                provider_with_zone.dns_service,
                'UpdateRecordSets',
                _update_record_sets,
            )

            # Large UPDATE_CHUNK_SIZE will put update in the same operation as delete & create
            provider_with_zone.apply(
                self._make_plan(
                    delete=[STUB_RECORDS['A']],
                    create=[STUB_RECORDS['MX']],
                    update=[
                        (STUB_RECORDS['CNAME_nodot'], STUB_RECORDS['CNAME'])
                    ],
                )
            )
            assert len(additions) == 1
            assert len(deletions) == 1

            self._compare_rset_batches(
                deletions[0], [STUB_RECORDS['A'], STUB_RECORDS['CNAME_nodot']]
            )
            self._compare_rset_batches(
                additions[0], [STUB_RECORDS['MX'], STUB_RECORDS['CNAME']]
            )

        def test_apply_big(self, monkeypatch, provider_with_zone):
            additions, deletions = [], []

            def _update_record_sets(request):
                additions.append(list(request.additions))
                deletions.append(list(request.deletions))
                return None

            provider_with_zone.UPDATE_CHUNK_SIZE = 2
            monkeypatch.setattr(
                provider_with_zone.dns_service,
                'UpdateRecordSets',
                _update_record_sets,
            )

            # Small UPDATE_CHUNK_SIZE will put update in different operation
            batch1_delete = [STUB_RECORDS['A'], STUB_RECORDS['CAA']]
            batch1_create = [STUB_RECORDS['AAAA'], STUB_RECORDS['CNAME']]
            batch2_delete = []
            batch2_create = [STUB_RECORDS['NS']]
            batch3_delete = [
                STUB_RECORDS['PTR_nodot'],
                STUB_RECORDS['SRV_nodot'],
            ]
            batch3_create = [STUB_RECORDS['PTR'], STUB_RECORDS['SRV']]

            provider_with_zone.apply(
                self._make_plan(
                    delete=batch1_delete + batch2_delete,
                    create=batch1_create + batch2_create,
                    update=[
                        (batch3_delete[0], batch3_create[0]),
                        (batch3_delete[1], batch3_create[1]),
                    ],
                )
            )
            assert len(additions) == 3
            assert len(deletions) == 3

            self._compare_rset_batches(deletions[0], batch1_delete)
            self._compare_rset_batches(additions[0], batch1_create)

            self._compare_rset_batches(deletions[1], batch2_delete)
            self._compare_rset_batches(additions[1], batch2_create)

            self._compare_rset_batches(deletions[2], batch3_delete)
            self._compare_rset_batches(additions[2], batch3_create)

        def test_apply_error(self, monkeypatch, provider_with_zone):
            def _update_record_sets(request):
                class RPCStateMock:
                    code = 1
                    details = "Some error"

                raise grpc.RpcError(RPCStateMock())

            monkeypatch.setattr(
                provider_with_zone.dns_service,
                'UpdateRecordSets',
                _update_record_sets,
            )

            with pytest.raises(YandexCloudException):
                provider_with_zone.apply(
                    self._make_plan([STUB_RECORDS['A']], [], [])
                )



from logging import getLogger

import grpc
import yandexcloud
from yandex.cloud.dns.v1.dns_zone_pb2 import RecordSet
from yandex.cloud.dns.v1.dns_zone_service_pb2 import (
    ListDnsZoneRecordSetsRequest,
    ListDnsZonesRequest,
    RecordSetDiff,
    UpdateRecordSetsMetadata,
    UpdateRecordSetsRequest,
)
from yandex.cloud.dns.v1.dns_zone_service_pb2_grpc import DnsZoneServiceStub

from octodns.idna import idna_decode
from octodns.provider.base import BaseProvider
from octodns.record import Record

from octodns_yandex.auth import _AuthMixin
from octodns_yandex.exception import YandexCloudException
from octodns_yandex.record import YandexCloudAnameRecord
from octodns_yandex.version import get_user_agent


def _aname_type_map(rset):
    rset.type = YandexCloudAnameRecord._type


def _txt_unescape(rset):
    # unescape value because octodns escaping breaks escaped dkim
    for i, value in enumerate(rset.data):
        rset.data[i] = value.replace('\\;', ';')


rset_transformers = {
    # Custom transformers for RecordSets from API
    'ANAME': _aname_type_map,
    'TXT': _txt_unescape,
}


def map_rset_to_octodns(provider, zone, lenient, rset):
    rset_transformers.get(rset.type, lambda *args: None)(rset)

    record_type = Record.registered_types().get(rset.type, None)
    if record_type is None:
        raise YandexCloudException(f"Unknown record type: {rset.type}")

    data = {'type': record_type._type, 'ttl': rset.ttl}

    values = record_type.parse_rdata_texts(rset.data)
    if len(values) == 1:
        data['value'] = values[0]
    else:
        data['values'] = values

    # name = zone.hostname_from_fqdn(rset.name)
    # fqdn = rset.name
    # if 0 < len(record_type.validate(name, fqdn, data)):
    #     data['octodns'] = {'lenient': True}

    return Record.new(
        zone,
        zone.hostname_from_fqdn(rset.name),
        data=data,
        source=provider,
        lenient=lenient,
    )


def map_octodns_to_rset(record: Record):
    values = record.data.get('values', record.data.get('value', []))
    values = values if isinstance(values, (list, tuple)) else [values]

    return RecordSet(
        name=record.fqdn,
        type=record._type.split('/')[-1],  # handle custom records
        ttl=record.ttl,
        data=[e.rdata_text for e in values],
    )


class YandexCloudProvider(_AuthMixin, BaseProvider):
    SUPPORTS_GEO = False
    SUPPORTS_DYNAMIC = False
    SUPPORTS_MULTIVALUE_PTR = True
    SUPPORTS_ROOT_NS = True  # Useless?
    SUPPORTS = {
        'A',
        'AAAA',
        'CAA',
        'CNAME',
        'MX',
        'NS',
        'PTR',
        # 'SOA',
        'SRV',
        # 'SVCB',
        # 'HTTPS',
        'TXT',
        YandexCloudAnameRecord._type,
    }

    UPDATE_CHUNK_SIZE = 1000

    prioritize_public = None
    auth_kwargs = dict()
    zone_ids_map = dict()

    sdk = None
    dns_service = None

    def __init__(
        self,
        id: str,
        folder_id: str,
        auth_type: str,
        prioritize_public=None,
        zone_ids_map=None,
        oauth_token=None,
        iam_token=None,
        sa_key_file=None,
        sa_key=None,
        *args,
        **kwargs,
    ):
        self.log = getLogger(f"YandexCloudProvider[{id}]")

        self.folder_id = folder_id
        self.prioritize_public = prioritize_public

        if isinstance(zone_ids_map, dict):
            self.zone_ids_map = zone_ids_map

        self.auth_kwargs = self.get_auth_kwargs(
            auth_type, oauth_token, iam_token, sa_key_file, sa_key
        )
        self.log.debug(
            '__init__: folder_id=%s auth_type=%s auth_kwargs=%s',
            self.folder_id,
            auth_type,
            self.auth_kwargs,
        )

        super().__init__(id, *args, **kwargs)

        self.sdk = yandexcloud.SDK(
            user_agent=get_user_agent(), **self.auth_kwargs
        )
        self.dns_service = self.sdk.client(DnsZoneServiceStub)

    def get_zone_id_by_name(self, zone_name):
        decoded_name = idna_decode(zone_name)
        mapped_id = self.zone_ids_map.get(
            decoded_name, self.zone_ids_map.get(zone_name, None)
        )
        if mapped_id is not None:
            self.log.debug(
                'get_zone_id_by_name: Found zone_name=%s in zone_ids_map',
                zone_name,
            )
            return mapped_id

        self.log.debug('get_zone_id_by_name: name=%s', decoded_name)

        # XXX: Will miss public zone if there is more than 1000 equally named internal zones
        zones = self.dns_service.List(
            ListDnsZonesRequest(
                folder_id=self.folder_id, filter=f'zone="{zone_name}"'
            )
        ).dns_zones

        if len(zones) < 1:
            self.log.debug('get_zone_id_by_name: No zones found')
            return None

        if len(zones) > 1 and self.prioritize_public is not None:
            if self.prioritize_public:
                public_zone = [
                    e for e in zones if e.HasField('public_visibility')
                ]
                if len(public_zone) > 0:
                    zones = public_zone
                    self.log.info(
                        'get_zone_id_by_name: Using public zone for zone_name=%s',
                        zone_name,
                    )
            else:
                zones = [e for e in zones if e.HasField('private_visibility')]
                self.log.info(
                    'get_zone_id_by_name: Searching for internal zones: zone_name=%s',
                    zone_name,
                )

        if len(zones) > 1:
            self.log.warning(
                "get_zone_id_by_name: Multiple zones found for zone_name=%s.\n"
                "Use 'prioritize_public' provider option to use public zones when present.\n"
                "Or use 'zone_ids_map' provider option to specify exact zone ids",
                zone_name,
            )

        zone = zones[0]
        self.log.info(
            'get_zone_id_by_name: Found zone_id=%s for zone_name=%s',
            zone.id,
            zone_name,
        )
        return zone.id

    def populate(self, zone, target=False, lenient=False):
        self.log.debug(
            'populate: name=%s, target=%s, lenient=%s',
            zone.name,
            target,
            lenient,
        )

        zone_id = self.get_zone_id_by_name(zone.name)
        if zone_id is None:
            self.log.info('populate: Zone not found')
            return False

        before = len(zone.records)
        done = False
        page_token = None
        while not done:
            resp = self.dns_service.ListRecordSets(
                ListDnsZoneRecordSetsRequest(
                    dns_zone_id=zone_id, page_token=page_token
                )
            )

            if resp.next_page_token:
                page_token = resp.next_page_token
            else:
                done = True

            for rset in resp.record_sets:
                if rset.type not in self.SUPPORTS | {'ANAME'}:
                    continue
                record = map_rset_to_octodns(self, zone, lenient, rset)
                zone.add_record(record, lenient=lenient)

        self.log.info('populate: found %s records', len(zone.records) - before)
        return True

    def _apply_rset_update(self, zone_id, create, delete):
        self.log.debug(
            'Applying changes:\n- Create: %s\n- Delete: %s', create, delete
        )

        try:
            operation = self.dns_service.UpdateRecordSets(
                UpdateRecordSetsRequest(
                    dns_zone_id=zone_id,
                    additions=[map_octodns_to_rset(e.new) for e in create],
                    deletions=[map_octodns_to_rset(e.existing) for e in delete],
                )
            )
            self.sdk.wait_operation_and_get_result(
                operation,
                response_type=RecordSetDiff,
                meta_type=UpdateRecordSetsMetadata,
            )
        except grpc.RpcError as e:
            state = e.args[0]
            raise YandexCloudException(
                f"API error: code={state.code}, details={state.details}"
            ) from e

    def _apply(self, plan):
        zone_name = plan.desired.name
        changes = plan.changes

        zone_id = self.get_zone_id_by_name(zone_name)
        if zone_id is None:
            raise YandexCloudException(
                'Zone not found (zone creation is not supported)'
            )

        self.log.debug(
            '_apply: zone_id=%s, zone_name=%s, len(changes)=%d',
            zone_id,
            zone_name,
            len(changes),
        )

        delete, create, update = [], [], []
        for change in changes:
            if change.new is None:
                delete.append(change)
            elif change.existing is None:
                create.append(change)
            else:
                update.append(change)

        # Try to process create & delete operations simultaneous in batches
        # Also, if there is enough free space in batch for all updates, then add them
        for i in range(
            0, max(len(delete), len(create)), self.UPDATE_CHUNK_SIZE
        ):
            create_chunk = create[i : i + self.UPDATE_CHUNK_SIZE]
            delete_chunk = delete[i : i + self.UPDATE_CHUNK_SIZE]

            max_len = max(len(create_chunk), len(delete_chunk))
            if (
                max_len < self.UPDATE_CHUNK_SIZE
                and max_len + len(update) <= self.UPDATE_CHUNK_SIZE
            ):
                create_chunk += update
                delete_chunk += update
                update = []
            self._apply_rset_update(zone_id, create_chunk, delete_chunk)

        # Process updates separately (if it was not done before)
        # API guarantees that deletions processed before additions
        for i in range(0, len(update), self.UPDATE_CHUNK_SIZE):
            chunk = update[i : i + self.UPDATE_CHUNK_SIZE]

            self._apply_rset_update(zone_id, chunk, chunk)



import pytest
import yandexcloud
from yandex.cloud.certificatemanager.v1.certificate_service_pb2 import (
    ListCertificatesResponse,
)
from yandex.cloud.certificatemanager.v1.certificate_service_pb2_grpc import (
    CertificateServiceStub,
)

from octodns.zone import Zone

from octodns_yandex import YandexCloudCMSource
from octodns_yandex.auth import AUTH_TYPE_METADATA
from octodns_yandex.exception import YandexCloudConfigException
from tests.fixtures import STUB_FOLDER_ID
from tests.fixtures.certificates import (
    HTTP_CHALLENGE_CERTIFICATE,
    IMPORTED_CERTIFICATE,
    STUB_CERTIFICATE_1,
    STUB_CERTIFICATE_2,
    STUB_IDNA_CERTIFICATE,
)


class StubSDK:
    def __init__(self, *args, **kwargs):
        pass

    def client(self, *args):
        class StubChannel:
            def unary_unary(self, *args, **kwargs):
                pass

        return CertificateServiceStub(StubChannel())

    def wait_operation_and_get_result(self, *args, **kwargs):
        pass


@pytest.fixture()
def disable_sdk(monkeypatch):
    monkeypatch.setattr(yandexcloud, 'SDK', StubSDK)
    return None


@pytest.fixture()
def provider(disable_sdk):
    return YandexCloudCMSource(
        "test", folder_id=STUB_FOLDER_ID, auth_type=AUTH_TYPE_METADATA
    )


class TestYandexCloudCMSource:
    def test_config_error(self, disable_sdk):
        with pytest.raises(YandexCloudConfigException):
            YandexCloudCMSource(
                "test",
                folder_id=STUB_FOLDER_ID,
                record_type='NS',
                auth_type=AUTH_TYPE_METADATA,
            )

    def test_process_certificate_managed(self, provider, monkeypatch):
        zone = Zone('example.com.', [])

        # Not matching
        provider.process_certificate(zone, STUB_IDNA_CERTIFICATE)
        assert len(zone.records) == 0

        # Not managed
        provider.process_certificate(zone, IMPORTED_CERTIFICATE)
        assert len(zone.records) == 0

        # Wrong challenge type
        provider.process_certificate(zone, HTTP_CHALLENGE_CERTIFICATE)
        assert len(zone.records) == 0

        # Domain + wildcard = challenge duplicate
        provider.process_certificate(zone, STUB_CERTIFICATE_1)
        assert len(zone.records) == 1

        record = zone.records.pop()
        assert record._type == 'CNAME'
        assert record.name == '_acme-challenge'
        assert (
            record.data['value'] == 'fpqahphei3aegootoh0a.cm.yandexcloud.net.'
        )
        assert record.data['ttl'] == provider.record_ttl
        zone.remove_record(record)

        # Two domains
        provider.process_certificate(zone, STUB_CERTIFICATE_2)
        assert len(zone.records) == 2

        records = zone.records
        for i in range(2):
            record = records.pop()
            assert record._type == 'CNAME'
            assert record.name in [
                '_acme-challenge.cdn',
                '_acme-challenge.cdn2',
            ]
            assert (
                record.data['value']
                == 'fpqco0ongail3feejee1.cm.yandexcloud.net.'
            )
            assert record.data['ttl'] == provider.record_ttl
            zone.remove_record(record)

        # Two domains, only one matches zone
        provider.record_type = 'TXT'
        provider.record_ttl = 300
        zone = Zone('cdn.example.com.', [])
        provider.process_certificate(zone, STUB_CERTIFICATE_2)
        assert len(zone.records) == 1

        record = zone.records.pop()
        assert record._type == 'TXT'
        assert record.name == '_acme-challenge'
        assert (
            record.data['value']
            == 'DIzWSC8sf8Sh2iMMO7chYopq8aAiNhcuxINGDKvONEA'
        )
        assert record.data['ttl'] == provider.record_ttl

        # Test idna
        zone = Zone('xn--e1aybc.xn--p1ai.', [])
        provider.process_certificate(zone, STUB_IDNA_CERTIFICATE)
        assert len(zone.records) == 1

        record = zone.records.pop()
        assert record._type == 'TXT'
        assert record.name == '_acme-challenge'
        assert record.data['values'] == [
            'ShsQrb08EUXyAmJ2UVGs5LPZuoOJEwTTmhQRM5LRq9T',
            'r4joRe9NgOIhyYLbNOCxdSPXNjnuU95kD1VbYbM0sNd',
        ]
        assert record.data['ttl'] == provider.record_ttl
        zone.remove_record(record)

    def test_populate(self, provider, monkeypatch):
        def _list(req):
            if req.page_token == '':
                return ListCertificatesResponse(
                    next_page_token='123', certificates=[STUB_CERTIFICATE_1]
                )
            else:
                return ListCertificatesResponse(
                    next_page_token='', certificates=[STUB_IDNA_CERTIFICATE]
                )

        monkeypatch.setattr(provider, 'process_certificate', lambda *args: None)
        monkeypatch.setattr(provider.cm_service, 'List', _list)

        zone = Zone('example.com.', [])
        provider.populate(zone)



from google.protobuf.timestamp_pb2 import Timestamp
from yandex.cloud.cdn.v1.resource_pb2 import (
    OriginProtocol,
    Resource,
    ResourceOptions,
    SSLCertificate,
    SSLCertificateCMData,
    SSLCertificateData,
    SSLCertificateStatus,
    SSLCertificateType,
)

from tests.fixtures import STUB_FOLDER_ID

STUB_CDN_RESOURCE_1 = Resource(
    id="bc8aePai2Xajaeghoo3b",
    folder_id=STUB_FOLDER_ID,
    cname="cdn.example.com",
    created_at=Timestamp(seconds=1711558651, nanos=963842000),
    updated_at=Timestamp(seconds=1711558659, nanos=106271000),
    active=True,
    options=ResourceOptions(
        edge_cache_settings=ResourceOptions.EdgeCacheSettings(
            enabled=True, default_value=345600
        ),
        browser_cache_settings=ResourceOptions.Int64Option(
            enabled=True, value=3600
        ),
        query_params_options=ResourceOptions.QueryParamsOptions(
            ignore_query_string=ResourceOptions.BoolOption(
                enabled=True, value=True
            )
        ),
        slice=ResourceOptions.BoolOption(enabled=True, value=True),
        compression_options=ResourceOptions.CompressionOptions(
            gzip_on=ResourceOptions.BoolOption(enabled=True)
        ),
        redirect_options=ResourceOptions.RedirectOptions(
            redirect_http_to_https=ResourceOptions.BoolOption(
                enabled=True, value=True
            )
        ),
        host_options=ResourceOptions.HostOptions(
            host=ResourceOptions.StringOption(
                enabled=True, value="origin.example.com"
            )
        ),
        static_headers=ResourceOptions.StringsMapOption(enabled=True),
        cors=ResourceOptions.StringsListOption(
            enabled=True, value=["domain1.example.com", "domain2.example.com"]
        ),
        stale=ResourceOptions.StringsListOption(
            enabled=True, value=["error", "updating"]
        ),
        allowed_http_methods=ResourceOptions.StringsListOption(
            enabled=True, value=["GET", "OPTIONS", "HEAD"]
        ),
        static_request_headers=ResourceOptions.StringsMapOption(enabled=True),
        ignore_cookie=ResourceOptions.BoolOption(enabled=True, value=True),
    ),
    origin_group_id=123456,
    origin_group_name="Stub group name 1",
    origin_protocol=OriginProtocol.HTTPS,
    ssl_certificate=SSLCertificate(
        type=SSLCertificateType.CM,
        status=SSLCertificateStatus.READY,
        data=SSLCertificateData(
            cm=SSLCertificateCMData(id="fpqzonu0haing3eegiet")
        ),
    ),
)
STUB_CDN_RESOURCE_2 = Resource(
    id="bc8ceeChozee1aCohcau",
    folder_id=STUB_FOLDER_ID,
    cname="cdn.example.com",
    created_at=Timestamp(seconds=1711558669, nanos=728117000),
    updated_at=Timestamp(seconds=1711558678, nanos=912906000),
    active=True,
    options=ResourceOptions(
        edge_cache_settings=ResourceOptions.EdgeCacheSettings(
            enabled=True, default_value=345600
        ),
        query_params_options=ResourceOptions.QueryParamsOptions(
            ignore_query_string=ResourceOptions.BoolOption(enabled=True)
        ),
        slice=ResourceOptions.BoolOption(enabled=True),
        compression_options=ResourceOptions.CompressionOptions(
            gzip_on=ResourceOptions.BoolOption(enabled=True, value=True)
        ),
        redirect_options=ResourceOptions.RedirectOptions(
            redirect_http_to_https=ResourceOptions.BoolOption(
                enabled=True, value=True
            )
        ),
        host_options=ResourceOptions.HostOptions(
            forward_host_header=ResourceOptions.BoolOption(
                enabled=True, value=True
            )
        ),
        static_headers=ResourceOptions.StringsMapOption(enabled=True),
        stale=ResourceOptions.StringsListOption(
            enabled=True, value=["error", "updating"]
        ),
        static_request_headers=ResourceOptions.StringsMapOption(enabled=True),
        ignore_cookie=ResourceOptions.BoolOption(enabled=True, value=True),
    ),
    secondary_hostnames=["cdn2.example.com."],
    origin_group_id=123457,
    origin_group_name="Stub group name 2",
    origin_protocol=OriginProtocol.HTTPS,
    ssl_certificate=SSLCertificate(
        type=SSLCertificateType.CM,
        status=SSLCertificateStatus.READY,
        data=SSLCertificateData(
            cm=SSLCertificateCMData(id="fpqnu9yohveig8ahm2je")
        ),
    ),
)



import json
import subprocess
import tempfile

import pytest

from octodns_yandex.auth import (
    AUTH_TYPE_IAM,
    AUTH_TYPE_METADATA,
    AUTH_TYPE_OAUTH,
    AUTH_TYPE_SA_KEY,
    AUTH_TYPE_YC_CLI,
    _AuthMixin,
)
from octodns_yandex.exception import YandexCloudConfigException

STUB_TOKEN = "token"
STUB_SA_KEY = {
    'id': 'id',
    'service_account_id': 'sa_id',
    'private_key': 'base64',
}


class TestYandexCloudAuth:
    class ProcessMock:
        def __init__(self, check_error=False, stdout=None):
            self.check_error = check_error
            self._stdout = stdout

        def check_returncode(self):
            if self.check_error:
                raise subprocess.CalledProcessError(
                    returncode=1, cmd='non_existent_file'
                )

        @property
        def stdout(self):
            return self._stdout

        @property
        def returncode(self):
            return 1

    @staticmethod
    def _get_auth_kwargs(auth_type, **kwargs):
        return _AuthMixin.get_auth_kwargs(
            auth_type,
            **{
                'oauth_token': None,
                'iam_token': None,
                'sa_key_file': None,
                'sa_key': None,
                **kwargs,
            }
        )

    def test_auth_oauth(self):
        auth_kwargs = self._get_auth_kwargs(
            AUTH_TYPE_OAUTH, oauth_token=STUB_TOKEN
        )
        assert auth_kwargs.keys() == {'token'}
        assert auth_kwargs['token'] == STUB_TOKEN

    def test_auth_iam(self):
        auth_kwargs = self._get_auth_kwargs(AUTH_TYPE_IAM, iam_token=STUB_TOKEN)
        assert auth_kwargs.keys() == {'iam_token'}
        assert auth_kwargs['iam_token'] == STUB_TOKEN

    def test_auth_metadata(self):
        auth_kwargs = self._get_auth_kwargs(AUTH_TYPE_METADATA)
        assert len(auth_kwargs) == 0

    def test_auth_service_account_key(self):
        # Auth using Service Account key, provided in file
        with tempfile.NamedTemporaryFile('w+') as fp:
            json.dump(STUB_SA_KEY, fp)
            fp.flush()
            auth_kwargs = self._get_auth_kwargs(
                AUTH_TYPE_SA_KEY, sa_key_file=fp.name
            )
            assert auth_kwargs.keys() == {'service_account_key'}
            assert auth_kwargs['service_account_key'] == STUB_SA_KEY

        with pytest.raises(
            YandexCloudConfigException, match=r".*non_existent_file.*"
        ):
            self._get_auth_kwargs(
                AUTH_TYPE_SA_KEY, sa_key_file='non_existent_file'
            )

        # Auth using Service Account key, provided in config
        auth_kwargs = self._get_auth_kwargs(
            AUTH_TYPE_SA_KEY, sa_key=STUB_SA_KEY
        )
        assert auth_kwargs.keys() == {'service_account_key'}
        assert auth_kwargs['service_account_key'] == STUB_SA_KEY

        # Wrong SA key format
        del STUB_SA_KEY['id']
        with pytest.raises(
            YandexCloudConfigException, match=r".*sa_key.*fields.*"
        ):
            self._get_auth_kwargs(AUTH_TYPE_SA_KEY, sa_key=STUB_SA_KEY)

    def test_auth_cli(self, monkeypatch):
        def _not_found(*args, **kwargs):
            raise FileNotFoundError()

        monkeypatch.setattr(
            subprocess,
            "run",
            lambda *args, **kwargs: self.ProcessMock(
                stdout=STUB_TOKEN.encode('utf-8')
            ),
        )
        auth_kwargs = self._get_auth_kwargs(AUTH_TYPE_YC_CLI)
        assert auth_kwargs.keys() == {'token'}
        assert auth_kwargs['token'] == STUB_TOKEN
        monkeypatch.undo()

        monkeypatch.setattr(subprocess, "run", _not_found)
        with pytest.raises(YandexCloudConfigException):
            self._get_auth_kwargs(AUTH_TYPE_YC_CLI)
        monkeypatch.undo()

        monkeypatch.setattr(
            subprocess,
            "run",
            lambda *args, **kwargs: self.ProcessMock(check_error=True),
        )
        with pytest.raises(YandexCloudConfigException):
            self._get_auth_kwargs(AUTH_TYPE_YC_CLI)
        monkeypatch.undo()

        # Unknown auth type
        with pytest.raises(YandexCloudConfigException):
            self._get_auth_kwargs('non_existent_type')



from octodns.record import Ipv6Value


def make_cursor_resp(page_token, **kwargs):
    return {'nextPageToken': page_token, **kwargs}


def make_paginated_resp(page, pages, per_page, total, **kwargs):
    return {
        'page': page,
        'pages': pages,
        'perPage': per_page,
        'total': total,
        **kwargs,
    }


STUB_ORG_1 = {
    'id': 1234,
    'name': 'example.com',
    'email': '',
    'phone': '',
    'fax': '',
    'language': 'ru',
    'subscriptionPlan': 'free',
}

STUB_ORG_2 = {
    'id': 1235,
    'name': 'example2.com',
    'email': '',
    'phone': '',
    'fax': '',
    'language': 'ru',
    'subscriptionPlan': 'free',
}


STUB_DOMAIN_1 = {
    'name': 'example.com',
    'country': 'ru',
    'mx': False,
    'delegated': False,
    'master': True,
    'verified': True,
    'status': {
        'name': 'example.com',
        'spf': {'match': False, 'value': 'v=spf1 mx a -all'},
        'mx': {'match': False, 'value': 'host=mx.example.com priority=0'},
        'ns': {'match': False, 'value': 'ns1.example.com\nns2.example.com'},
        'dkim': {
            'match': True,
            'value': 'v=DKIM1; k=rsa; t=s; p=MIGfMA0GCSqGSIb3DQEBAQUAA4GNADCBiQKBgQDK1bS'
            '7mNooVu1/GB8FKDwp3gxjzFCUeAz60IM7KOIsOhqFiVqkvr7zV35C5FLwsx6buJbpl'
            'AcgTjvLzKwgDfldUsStjkmIxMHNFY9uXF/+4Go/wXLpe5SwAsXdmAe7IOeu27Issyr'
            'PJ74WYE+3iAeByChXYIcY7oHEthcfwlQk8wIDAQAB',
        },
        'lastCheck': '2024-03-23T07:39:41.945Z',
        'lastAdded': '2024-03-20T14:23:46.931Z',
    },
}
STUB_DOMAIN_2 = {
    'name': 'ya360.example.com',
    'country': 'ru',
    'mx': True,
    'delegated': True,
    'master': False,
    'verified': True,
}

# Every value should be a set of entries resulting in single Record
# Value which key equals DNS record type should not be lenient
STUB_ENTRIES = {
    'A': [
        {
            'recordId': 1,
            'type': 'A',
            'name': 'a',
            'ttl': 600,
            'address': '127.0.0.1',
        },
        {
            'recordId': 2,
            'type': 'A',
            'name': 'a',
            'ttl': 600,
            'address': '127.0.0.2',
        },
    ],
    'A_single': [
        {
            'recordId': 3,
            'type': 'A',
            'name': 'a1',
            'ttl': 600,
            'address': '127.0.0.1',
        }
    ],
    'AAAA': [
        {
            'recordId': 4,
            'type': 'AAAA',
            'name': 'aaaa',
            'ttl': 300,
            'address': str(
                Ipv6Value('2001:0db8:11a3:09d7:1f34:8a2e:07a0:765d')
            ),
        },
        {
            'recordId': 5,
            'type': 'AAAA',
            'name': 'aaaa',
            'ttl': 350,
            'address': str(
                Ipv6Value('2001:0db8:11a3:09d7:1f34:8a2e:07a0:766d')
            ),
        },
    ],
    'AAAA_triple': [
        {
            'recordId': 6,
            'type': 'AAAA',
            'name': 'aaaa3',
            'ttl': 300,
            'address': str(
                Ipv6Value('2001:0db8:11a3:09d7:1f34:8a2e:07a0:765d')
            ),
        },
        {
            'recordId': 7,
            'type': 'AAAA',
            'name': 'aaaa3',
            'ttl': 350,
            'address': str(
                Ipv6Value('2001:0db8:11a3:09d7:1f34:8a2e:07a0:766d')
            ),
        },
        {
            'recordId': 8,
            'type': 'AAAA',
            'name': 'aaaa3',
            'ttl': 350,
            'address': str(
                Ipv6Value('2001:0db8:11a3:09d7:1f34:8a2e:07a0:767d')
            ),
        },
    ],
    'AAAA_second_record': [
        {
            'recordId': 9,
            'type': 'AAAA',
            'name': 'a',
            'ttl': 600,
            'address': '::1',
        }
    ],
    'CNAME': [
        {
            'recordId': 10,
            'type': 'CNAME',
            'name': 'cname',
            'ttl': 600,
            'target': 'example.com.',
        }
    ],
    'CNAME_mail': [
        {
            'recordId': 11,
            'type': 'CNAME',
            'name': 'mail',
            'ttl': 21600,
            'target': 'domain.mail.yandex.net.',
        }
    ],
    'MX': [
        {
            'recordId': 12,
            'type': 'MX',
            'name': '@',
            'ttl': 21600,
            'exchange': 'mx.yandex.net.',
            'preference': 10,
        }
    ],
    'TXT': [
        {
            'recordId': 13,
            'type': 'TXT',
            'name': 'txt',
            'ttl': 600,
            'text': 'Hello world',
        }
    ],
    'TXT_dkim': [
        {
            'recordId': 14,
            'type': 'TXT',
            'name': 'mail._domainkey',
            'ttl': 21600,
            'text': 'v=DKIM1; k=rsa\\\\; t=s; p=MIGfMA0GCSqGSIb3DQEBAQUAA4GNADCBiQ'
            'KBgQDK1bS7mNooVu1/GB8FKDwp3gxjzFCUeAz60IM7KOIsOhqFiVqkvr7zV35C5FLwsx6'
            'buJbplAcgTjvLzKwgDfldUsStjkmIxMHNFY9uXF/+4Go/wXLpe5SwAsXdmAe7IOeu27Is'
            'syrPJ74WYE+3iAeByChXYIcY7oHEthcfwlQk8wIDAQAB',
        }
    ],
    'TXT_spf': [
        {
            'recordId': 15,
            'type': 'TXT',
            'name': '@',
            'ttl': 21600,
            'text': 'v=spf1 redirect=_spf.yandex.net',
        }
    ],
    'SRV': [
        {
            'recordId': 16,
            'type': 'SRV',
            'name': '_service._tcp',
            'ttl': 600,
            'target': 'backend.example.com.',
            'port': 8080,
            'priority': 10,
            'weight': 70,
        }
    ],
    'NS': [
        {
            'recordId': 17,
            'type': 'NS',
            'name': 'ns',
            'ttl': 600,
            'target': 'ns1.example.com.',
        },
        {
            'recordId': 18,
            'type': 'NS',
            'name': 'ns',
            'ttl': 600,
            'target': 'ns2.example.com.',
        },
    ],
    'CAA': [
        {
            'recordId': 19,
            'type': 'CAA',
            'name': 'caa',
            'ttl': 600,
            'flag': 0,
            'tag': 'issue',
            'value': 'ca.example.net',
        }
    ],
}



import pytest
import yandexcloud
from yandex.cloud.cdn.v1.resource_service_pb2 import (
    GetProviderCNameResponse,
    ListResourcesResponse,
)
from yandex.cloud.cdn.v1.resource_service_pb2_grpc import ResourceServiceStub

from octodns.record import Record
from octodns.zone import Zone

from octodns_yandex import YandexCloudCDNSource
from octodns_yandex.auth import AUTH_TYPE_METADATA
from tests.fixtures import STUB_FOLDER_ID
from tests.fixtures.cdn_resources import (
    STUB_CDN_RESOURCE_1,
    STUB_CDN_RESOURCE_2,
)


class StubSDK:
    def __init__(self, *args, **kwargs):
        pass

    def client(self, *args):
        class StubChannel:
            def unary_unary(self, *args, **kwargs):
                pass

        return ResourceServiceStub(StubChannel())

    def wait_operation_and_get_result(self, *args, **kwargs):
        pass


@pytest.fixture()
def disable_sdk(monkeypatch):
    monkeypatch.setattr(yandexcloud, 'SDK', StubSDK)
    return None


@pytest.fixture()
def provider(disable_sdk):
    return YandexCloudCDNSource(
        "test", folder_id=STUB_FOLDER_ID, auth_type=AUTH_TYPE_METADATA
    )


class TestYandexCloudCDNSource:
    def test_get_provider_cname(self, provider, monkeypatch):
        monkeypatch.setattr(
            provider.cdn_service,
            'GetProviderCName',
            lambda *args: GetProviderCNameResponse(cname='test.com'),
        )
        assert provider.get_provider_cname() == 'test.com.'
        monkeypatch.undo()

        def raise_exc(*args):
            raise Exception("Failed")

        monkeypatch.setattr(provider.cdn_service, 'GetProviderCName', raise_exc)
        assert provider.get_provider_cname() == 'test.com.'

    def test_get_provider_cname_with_dot(self, provider, monkeypatch):
        monkeypatch.setattr(
            provider.cdn_service,
            'GetProviderCName',
            lambda *args: GetProviderCNameResponse(cname='test.com.'),
        )
        assert provider.get_provider_cname() == 'test.com.'

    def test_process_resource(self, provider, monkeypatch):
        zone = Zone('example.com.', [])

        CDN_PROVIDER_CNAME = 'test.com.'
        monkeypatch.setattr(
            provider, 'get_provider_cname', lambda *args: CDN_PROVIDER_CNAME
        )

        MAIN_RECORD = Record.new(
            zone,
            'cdn',
            data={
                'type': 'CNAME',
                'ttl': provider.record_ttl,
                'value': CDN_PROVIDER_CNAME,
            },
        )
        SECONDARY_RECORD = Record.new(
            zone,
            'cdn2',
            data={
                'type': 'CNAME',
                'ttl': provider.record_ttl,
                'value': CDN_PROVIDER_CNAME,
            },
        )

        provider.process_resource(zone, STUB_CDN_RESOURCE_1)
        assert len(zone.records) == 1
        assert zone.records.pop() == MAIN_RECORD
        zone.remove_record(MAIN_RECORD)

        provider.process_resource(zone, STUB_CDN_RESOURCE_2)
        assert len(zone.records) == 2
        assert next((r for r in zone.records if r.name == 'cdn')) == MAIN_RECORD
        assert (
            next((r for r in zone.records if r.name == 'cdn2'))
            == SECONDARY_RECORD
        )

    def test_populate(self, provider, monkeypatch):
        def _list(req):
            if req.page_token == '':
                return ListResourcesResponse(
                    next_page_token='123', resources=[STUB_CDN_RESOURCE_1]
                )
            else:
                return ListResourcesResponse(
                    next_page_token='', resources=[STUB_CDN_RESOURCE_2]
                )

        monkeypatch.setattr(provider, 'process_resource', lambda *args: None)
        monkeypatch.setattr(provider.cdn_service, 'List', _list)

        zone = Zone('example.com.', [])
        provider.populate(zone)



import functools
import itertools
from logging import getLogger

import requests

from octodns.idna import idna_encode
from octodns.provider import ProviderException
from octodns.provider.base import BaseProvider
from octodns.record import Create, Delete, Record

from octodns_yandex.version import get_base_user_agent


class Yandex360Exception(ProviderException):
    pass


class Yandex360ApiException(ProviderException):
    def __init__(self, resp):
        super(Yandex360ApiException, self).__init__(
            f"Error code: {resp.status_code}, body: {resp.text}"
        )


def _octodns_name(name):
    return name if name != '@' else ''


def _ya360_name(name):
    return name if name != '' else '@'


def map_entries_to_records(provider, zone, lenient, entries):
    def _keyfunc(x):
        return x['name'], x['type']

    records = []
    entries = sorted(entries, key=_keyfunc)
    for k, entry_group in itertools.groupby(entries, _keyfunc):
        name, type = k
        entry_group = list(entry_group)

        record_type = Record.registered_types().get(type, None)
        if record_type is None:
            raise Yandex360Exception(f"Unknown record type: {type}")

        # A little helper to construct lists with singular octodns record values
        def make_value(*args, **kwargs):
            value_type = record_type._value_type
            return [value_type(kwargs) if kwargs else value_type(*args)]

        values = []
        for entry in entry_group:
            if type in {'A', 'AAAA'}:
                values += make_value(entry['address'])
            elif type in {'CNAME', 'NS'}:
                values += make_value(entry['target'])
            elif type == 'TXT':
                # Hard escape to meet octodns requirements. We will unescape on apply
                values += make_value(entry['text'].replace(';', '\\;'))
            elif type == 'MX':
                values += make_value(
                    preference=int(entry['preference']),
                    exchange=entry['exchange'],
                )
            elif type == 'SRV':
                values += make_value(
                    priority=int(entry['priority']),
                    weight=int(entry['weight']),
                    port=int(entry['port']),
                    target=entry['target'],
                )
            elif type == 'CAA':
                values += make_value(
                    flags=int(entry['flag']),
                    tag=entry['tag'],
                    value=entry['value'],
                )
            else:
                raise Yandex360Exception(f"Unknown record type: {type}")

        data = {
            'type': type,
            # We can't really determine unified TTL
            'ttl': entry_group[0]['ttl'],
        }
        if len(values) == 1:
            data['value'] = values[0]
        else:
            data['values'] = values

        records.append(
            Record.new(
                zone,
                _octodns_name(name),
                data=data,
                source=provider,
                lenient=lenient,
            )
        )

    return records


def map_record_to_entries(zone, record: Record):
    type = record._type
    name = zone.hostname_from_fqdn(record.name)

    values = record.data.get('values', record.data.get('value', []))
    values = values if isinstance(values, (list, tuple)) else [values]

    # A little helper to construct lists with singular entries which have similar fields
    make_entry = functools.partial(
        lambda **kwargs: [dict(**kwargs)],
        name=_ya360_name(name),
        type=type,
        ttl=int(record.ttl),
    )

    entries = []
    for value in values:
        if type in {'A', 'AAAA'}:
            entries += make_entry(address=value)
        elif type in {'CNAME', 'NS'}:
            entries += make_entry(target=value)
        elif type == 'TXT':
            # NOTE: Yandex 360 is escaping semicolons properly
            # XXX: Yandex 360 is doubling backslashes
            entries += make_entry(
                text=value.replace('\\;', ';').replace('\\\\', '\\')
            )
        elif type == 'MX':
            entries += make_entry(
                preference=int(value.preference), exchange=value.exchange
            )
        elif type == 'SRV':
            entries += make_entry(
                priority=int(value.priority),
                weight=int(value.weight),
                port=int(value.port),
                target=value.target,
            )
        elif type == 'CAA':
            entries += make_entry(
                flag=int(value.flags), tag=value.tag, value=value.value
            )
        else:
            raise Yandex360Exception(f"Unknown record type: {type}")

    return entries


# API reference:
#  https://yandex.ru/dev/api360/doc/ref/OrganizationsService/OrganizationsService_List.html
#  https://yandex.ru/dev/api360/doc/ref/DomainService/DomainService_List.html
#  https://yandex.ru/dev/api360/doc/ref/DomainDNSService.html


class Yandex360Provider(BaseProvider):
    SUPPORTS_GEO = False
    SUPPORTS_DYNAMIC = False
    SUPPORTS = {'A', 'AAAA', 'CNAME', 'MX', 'TXT', 'SRV', 'NS', 'CAA'}

    TIMEOUT = 15
    API_BASE = 'https://api360.yandex.net'

    _oauth_token = None

    def __init__(self, id, oauth_token, *args, **kwargs):
        self.log = getLogger(f"Yandex360Provider[{id}]")

        self._oauth_token = oauth_token

        self.log.debug('__init__: oauth_token=%s', self._oauth_token)

        super().__init__(id, *args, **kwargs)

        self._session = requests.Session()
        self._session.headers.update(
            {
                'Authorization': f"OAuth {self._oauth_token}",
                'User-Agent': get_base_user_agent(),
            }
        )

    def make_request(
        self, method, url, data=None, params=None, expected_code=200
    ):
        resp = self._session.request(
            method,
            f"{self.API_BASE}{url}",
            params=params,
            json=data,
            timeout=self.TIMEOUT,
        )
        if resp.status_code != expected_code:
            raise Yandex360ApiException(resp)
        return resp.json()

    def list_orgs(self, page_token=None):
        return self.make_request(
            'GET',
            '/directory/v1/org',
            params={'pageSize': 100, 'pageToken': page_token},
        )

    def list_domains(self, org_id, page=1):
        return self.make_request(
            'GET',
            f"/directory/v1/org/{org_id}/domains",
            params={'perPage': 10, 'page': page},
        )

    def list_dns_records(self, org_id, domain, page=1):
        return self.make_request(
            'GET',
            f"/directory/v1/org/{org_id}/domains/{domain}/dns",
            params={'perPage': 50, 'page': page},
        )

    def create_dns_record(self, org_id, domain, data):
        self.log.debug(
            "API Create org_id=%s, domain=%s, data=%s", org_id, domain, data
        )
        return self.make_request(
            'POST',
            f"/directory/v1/org/{org_id}/domains/{domain}/dns",
            data=data,
        )

    def update_dns_record(self, org_id, domain, record_id, data):
        self.log.debug(
            "API Update org_id=%s, domain=%s, record_id+%s, data=%s",
            org_id,
            domain,
            record_id,
            data,
        )
        return self.make_request(
            'POST',
            f"/directory/v1/org/{org_id}/domains/{domain}/dns/{record_id}",
            data=data,
        )

    def delete_dns_record(self, org_id, domain, record_id):
        self.log.debug(
            "API Delete org_id=%s, domain=%s, record_id=%s",
            org_id,
            domain,
            record_id,
        )
        return self.make_request(
            'DELETE',
            f"/directory/v1/org/{org_id}/domains/{domain}/dns/{record_id}",
        )

    def find_org_id_for_domain(self, domain_name):
        orgs_done = False
        orgs_page_token = None
        while not orgs_done:
            orgs_resp = self.list_orgs(page_token=orgs_page_token)

            if orgs_resp['nextPageToken']:
                orgs_page_token = orgs_resp['nextPageToken']
            else:
                orgs_done = True

            for org in orgs_resp['organizations']:
                org_id = org['id']

                domains_page, domains_pages = 1, 1
                while domains_page <= domains_pages:
                    domains_resp = self.list_domains(org_id, page=domains_page)

                    domains_pages = domains_resp['pages']
                    domains_page += 1

                    for domain in domains_resp['domains']:
                        if domain['name'] != domain_name:
                            continue
                        self.log.info(
                            'find_org_id_for_domain: Found org_id=%s for domain_name=%s',
                            org_id,
                            domain_name,
                        )
                        return org_id

        return None

    def collect_zone_entries(self, org_id, domain_name):
        entries = []

        records_page, records_pages = 1, 1
        while records_page <= records_pages:
            domains_resp = self.list_dns_records(
                org_id, domain_name, page=records_page
            )

            records_pages = domains_resp['pages']
            records_page += 1

            entries += domains_resp['records']

        return entries

    def populate(self, zone, target=False, lenient=False):
        self.log.debug(
            'populate: name=%s, target=%s, lenient=%s',
            zone.name,
            target,
            lenient,
        )

        domain_name = idna_encode(zone.name).rstrip('.')
        org_id = self.find_org_id_for_domain(domain_name)
        if org_id is None:
            self.log.info('populate: Zone not found')
            return False

        before = len(zone.records)
        entries = self.collect_zone_entries(org_id, domain_name)
        for record in map_entries_to_records(self, zone, lenient, entries):
            zone.add_record(record, lenient=lenient)

        self.log.info('populate: found %s records', len(zone.records) - before)

        return True

    def _apply(self, plan):
        zone = plan.desired
        changes = plan.changes

        domain_name = idna_encode(zone.name).rstrip('.')
        org_id = self.find_org_id_for_domain(domain_name)
        if org_id is None:
            raise Yandex360Exception("Zone not found")

        self.log.debug(
            '_apply: org_id=%s, domain_name=%s, len(changes)=%d',
            org_id,
            domain_name,
            len(changes),
        )

        delete, create, update = [], [], []
        records_to_search = {}
        for change in changes:
            if change.existing is None:
                create.append(change)
            else:
                _key = (
                    change.existing._type,
                    _ya360_name(change.existing.name),
                )
                records_to_search[_key] = []
                if change.new is None:
                    delete.append(change)
                elif change.existing._type == 'CAA':
                    # XXX: CAA update is broken
                    delete.append(Delete(change.existing))
                    create.append(Create(change.existing))
                else:
                    update.append(change)

        # Search for record_ids by (type, name) tuples
        entries = self.collect_zone_entries(org_id, domain_name)
        for entry in entries:
            _key = (entry['type'], entry['name'])
            e = records_to_search.get(_key, None)
            if e is None:
                continue
            records_to_search[_key].append(entry['recordId'])

        # Delete found records
        for change in delete:
            _key = (change.existing._type, _ya360_name(change.existing.name))
            for record_id in records_to_search.get(_key, []):
                self.delete_dns_record(org_id, domain_name, record_id)

        # Create new records
        for change in create:
            for entry in map_record_to_entries(zone, change.new):
                self.create_dns_record(org_id, domain_name, entry)

        # Apply changes: update (if possible) or create/delete
        for change in update:
            _key = (change.existing._type, _ya360_name(change.existing.name))
            it = iter(records_to_search.get(_key, []))

            # Update entries while there is some or create new ones
            for entry in map_record_to_entries(zone, change.new):
                record_id = next(it, None)
                if record_id is not None:
                    self.update_dns_record(
                        org_id, domain_name, record_id, entry
                    )
                else:
                    self.create_dns_record(org_id, domain_name, entry)

            # Delete additional entries (if any)
            for record_id in it:
                self.delete_dns_record(org_id, domain_name, record_id)



import functools

import pytest

from octodns.provider.plan import Plan
from octodns.record import Create, Delete, PtrRecord, Update
from octodns.zone import Zone

from octodns_yandex import Yandex360Provider
from octodns_yandex.yandex360_provider import (
    Yandex360ApiException,
    Yandex360Exception,
    map_entries_to_records,
    map_record_to_entries,
)
from tests.fixtures.ya360 import (
    STUB_DOMAIN_1,
    STUB_DOMAIN_2,
    STUB_ENTRIES,
    STUB_ORG_1,
    STUB_ORG_2,
    make_cursor_resp,
    make_paginated_resp,
)


@pytest.fixture()
def provider(monkeypatch):
    return Yandex360Provider('test', 'token')


class TestYandex360Provider:
    def test_make_request(self, provider, monkeypatch):
        class MockSession:
            status_code = 200
            text = "Test body"

            @staticmethod
            def request(*args, **kwargs):
                return MockSession

            @staticmethod
            def json():
                return {}

        monkeypatch.setattr(provider, "_session", MockSession)

        provider.make_request('GET', '/test', expected_code=200)
        with pytest.raises(
            Yandex360ApiException, match=f".*{MockSession.text}.*"
        ):
            provider.make_request('GET', '/test', expected_code=204)

    def test_api_urls(self, provider, monkeypatch):
        last_method, last_url = None, None

        def save_args(method, url, *args, **kwargs):
            nonlocal last_method, last_url
            last_method = method
            last_url = url

        monkeypatch.setattr(provider, 'make_request', save_args)

        provider.list_orgs(page_token=None)
        assert (last_method, last_url) == ('GET', '/directory/v1/org')

        provider.list_domains(123, page=1)
        assert (last_method, last_url) == (
            'GET',
            "/directory/v1/org/123/domains",
        )

        provider.list_dns_records(123, 'example.com')
        assert (last_method, last_url) == (
            'GET',
            '/directory/v1/org/123/domains/example.com/dns',
        )

        provider.create_dns_record(123, 'example.com', {})
        assert (last_method, last_url) == (
            'POST',
            '/directory/v1/org/123/domains/example.com/dns',
        )

        provider.update_dns_record(123, 'example.com', 12345, {})
        assert (last_method, last_url) == (
            'POST',
            '/directory/v1/org/123/domains/example.com/dns/12345',
        )

        provider.delete_dns_record(123, 'example.com', 12345)
        assert (last_method, last_url) == (
            'DELETE',
            '/directory/v1/org/123/domains/example.com/dns/12345',
        )

    def test_find_org_not_found(self, provider, monkeypatch):
        monkeypatch.setattr(
            provider,
            'list_orgs',
            lambda page_token: make_cursor_resp('', organizations=[]),
        )

        found_org_id = provider.find_org_id_for_domain(STUB_DOMAIN_1['name'])
        assert found_org_id is None

    def test_find_org_multiple_pages(self, provider, monkeypatch):
        def _list_orgs(page_token):
            if not page_token:
                orgs = [STUB_ORG_1]
            else:
                orgs = [STUB_ORG_2]
            return make_cursor_resp(
                '2' if not page_token else '', organizations=orgs
            )

        def _list_domains(org_id, page):
            if org_id == STUB_ORG_1['id']:
                domains = []
            elif page == 1:
                domains = [STUB_DOMAIN_1]
            else:
                domains = [STUB_DOMAIN_2]
            return make_paginated_resp(page, 2, 1, 2, domains=domains)

        monkeypatch.setattr(provider, 'list_orgs', _list_orgs)
        monkeypatch.setattr(provider, 'list_domains', _list_domains)

        found_org_id = provider.find_org_id_for_domain(STUB_DOMAIN_2['name'])
        assert found_org_id == STUB_ORG_2['id']

    def test_mapping_supported(self):
        zone = Zone('example.com.', [])

        for type in Yandex360Provider.SUPPORTS:
            entries = STUB_ENTRIES[type.split('/')[-1]]
            records = map_entries_to_records(None, zone, False, entries)
            assert len(records) == 1

            ttls = set(map(lambda x: x['ttl'], entries))

            recovered_entries = map_record_to_entries(zone, records[0])
            assert len(recovered_entries) == len(entries)

            for i in range(len(recovered_entries)):
                copy = entries[i].copy()
                del copy['recordId'], copy['ttl']
                ttl = recovered_entries[i].pop('ttl')

                assert recovered_entries[i] == copy
                assert ttl in ttls

    def test_mapping_unknown(self):
        zone = Zone('example.com.', [])

        # Not known to octodns
        with pytest.raises(Yandex360Exception):
            map_entries_to_records(
                None,
                zone,
                False,
                [
                    {
                        'recordId': 8,
                        'type': 'NON_EXISTENT_TYPE',
                        'name': '@',
                        'ttl': 600,
                    }
                ],
            )

        # Not known to yandex 360
        with pytest.raises(Yandex360Exception):
            map_entries_to_records(
                None,
                zone,
                False,
                [{'recordId': 8, 'type': 'PTR', 'name': '@', 'ttl': 600}],
            )

        with pytest.raises(Yandex360Exception):
            map_record_to_entries(
                zone,
                PtrRecord(zone, "test", {'ttl': 600, 'value': "example.com."}),
            )

    def test_collect_zone_entries(self, provider, monkeypatch):
        def _list_dns_records(org_id, domain_name, page):
            if page == 1:
                records = STUB_ENTRIES['A']
            else:
                records = STUB_ENTRIES['MX']
            return make_paginated_resp(page, 2, 1, 2, records=records)

        monkeypatch.setattr(provider, 'list_dns_records', _list_dns_records)

        entries = provider.collect_zone_entries(
            STUB_ORG_1['id'], STUB_DOMAIN_1['name']
        )
        assert entries == STUB_ENTRIES['A'] + STUB_ENTRIES['MX']

    def test_populate_not_found(self, provider, monkeypatch):
        zone = Zone('example.com.', [])

        monkeypatch.setattr(
            provider, 'find_org_id_for_domain', lambda domain_name: None
        )

        result = provider.populate(zone)
        assert not result

    def test_populate_found(self, provider, monkeypatch):
        TEST_ENTRIES = list(
            functools.reduce(lambda acc, x: acc + x, STUB_ENTRIES.values(), [])
        )
        zone = Zone('example.com.', [])

        monkeypatch.setattr(
            provider,
            'find_org_id_for_domain',
            lambda domain_name: STUB_ORG_1['id'],
        )
        monkeypatch.setattr(
            provider,
            'collect_zone_entries',
            lambda org_id, domain_name: TEST_ENTRIES,
        )

        result = provider.populate(zone, lenient=True)
        assert result

        assert len(zone.records) == len(STUB_ENTRIES.keys())

    @staticmethod
    def _make_plan(create, delete, update):
        zone = Zone('example.com.', [])
        return Plan(
            existing=None,
            desired=zone,
            changes=[
                *map(Create, map_entries_to_records(None, zone, True, create)),
                *map(Delete, map_entries_to_records(None, zone, True, delete)),
                *map(
                    lambda x: Update(*x),
                    map(
                        lambda x: (
                            map_entries_to_records(None, zone, True, x[0])[0],
                            map_entries_to_records(None, zone, True, x[1])[0],
                        ),
                        update,
                    ),
                ),
            ],
            exists=False,
        )

    def test_apply_not_found(self, provider, monkeypatch):
        monkeypatch.setattr(
            provider, 'find_org_id_for_domain', lambda domain_name: None
        )
        with pytest.raises(Yandex360Exception):
            provider.apply(self._make_plan([], [], []))

    def test_apply(self, provider, monkeypatch):
        TEST_ENTRIES = functools.reduce(
            lambda acc, x: acc + x, STUB_ENTRIES.values(), []
        )
        created_tuples, deleted_ids, updated_ids = [], [], []

        monkeypatch.setattr(
            provider,
            'find_org_id_for_domain',
            lambda domain_name: STUB_ORG_1['id'],
        )
        monkeypatch.setattr(
            provider,
            'collect_zone_entries',
            lambda org_id, domain_name: TEST_ENTRIES,
        )

        monkeypatch.setattr(
            provider,
            'delete_dns_record',
            lambda org_id, domain, record_id: deleted_ids.append(record_id),
        )
        monkeypatch.setattr(
            provider,
            'create_dns_record',
            lambda org_id, domain, data: created_tuples.append(
                (data['type'], data['name'])
            ),
        )
        monkeypatch.setattr(
            provider,
            'update_dns_record',
            lambda org_id, domain, record_id, data: updated_ids.append(
                record_id
            ),
        )

        provider.apply(
            self._make_plan(
                create=STUB_ENTRIES['A'] + STUB_ENTRIES['CNAME'],
                delete=STUB_ENTRIES['MX'] + STUB_ENTRIES['SRV'],
                update=[
                    (
                        STUB_ENTRIES['CNAME'],
                        STUB_ENTRIES['CNAME_mail'],
                    ),  # Basic update
                    (
                        STUB_ENTRIES['A'],
                        STUB_ENTRIES['A_single'],
                    ),  # Remove one entry
                    (
                        STUB_ENTRIES['AAAA'],
                        STUB_ENTRIES['AAAA_triple'],
                    ),  # Add one entry
                    # CAA update is broken, so just make sure it does not invoke an update operation
                    (STUB_ENTRIES['CAA'], STUB_ENTRIES['CAA']),
                ],
            )
        )

        # Check applied operations
        # NOTE: octodns could reorder changes before applying
        assert created_tuples == list(
            map(
                lambda x: (x['type'], x['name']),
                STUB_ENTRIES['A']
                + STUB_ENTRIES['CNAME']
                + STUB_ENTRIES['CAA']
                + [STUB_ENTRIES['AAAA_triple'][2]],
            )
        )
        assert deleted_ids == list(
            map(
                lambda x: x['recordId'],
                STUB_ENTRIES['MX']
                + STUB_ENTRIES['SRV']
                + STUB_ENTRIES['CAA']
                + [STUB_ENTRIES['A'][1]],
            )
        )
        assert updated_ids == list(
            map(
                lambda x: x['recordId'],
                [STUB_ENTRIES['A'][0]]
                + STUB_ENTRIES['AAAA']
                + STUB_ENTRIES['CNAME'],
            )
        )



from logging import getLogger

import yandexcloud
from yandex.cloud.cdn.v1.resource_service_pb2 import (
    GetProviderCNameRequest,
    ListResourcesRequest,
)
from yandex.cloud.cdn.v1.resource_service_pb2_grpc import ResourceServiceStub

from octodns.record import Record
from octodns.source.base import BaseSource

from octodns_yandex.auth import _AuthMixin
from octodns_yandex.version import get_user_agent


class YandexCloudCDNSource(_AuthMixin, BaseSource):
    SUPPORTS_GEO = False
    SUPPORTS = {'CNAME'}

    _provider_cname = None

    def __init__(
        self,
        id,
        folder_id: str,
        auth_type: str,
        record_ttl=3600,
        oauth_token=None,
        iam_token=None,
        sa_key_file=None,
        sa_key=None,
        *args,
        **kwargs,
    ):
        self.log = getLogger(f"YandexCloudCDNSource[{id}]")

        self.folder_id = folder_id
        self.record_ttl = record_ttl

        self.auth_kwargs = self.get_auth_kwargs(
            auth_type, oauth_token, iam_token, sa_key_file, sa_key
        )
        self.log.debug(
            '__init__: folder_id=%s auth_type=%s auth_kwargs=%s',
            self.folder_id,
            auth_type,
            self.auth_kwargs,
        )

        super().__init__(id, *args, **kwargs)

        self.sdk = yandexcloud.SDK(
            user_agent=get_user_agent(), **self.auth_kwargs
        )
        self.cdn_service = self.sdk.client(ResourceServiceStub)

    def get_provider_cname(self):
        if not self._provider_cname:
            self._provider_cname = self.cdn_service.GetProviderCName(
                GetProviderCNameRequest(folder_id=self.folder_id)
            ).cname
            if self._provider_cname[-1] != '.':
                self._provider_cname += '.'

        return self._provider_cname

    def process_resource(self, zone, resource, lenient=False):
        matched_domains = filter(
            lambda domain: zone.owns('CNAME', domain),
            [resource.cname, *resource.secondary_hostnames],
        )

        for fqdn in matched_domains:
            if fqdn[-1] != '.':
                fqdn = f'{fqdn}.'

            zone.add_record(
                Record.new(
                    zone,
                    zone.hostname_from_fqdn(fqdn),
                    data={
                        'type': 'CNAME',
                        'ttl': self.record_ttl,
                        'value': self.get_provider_cname(),
                    },
                    source=self,
                    lenient=lenient,
                )
            )

    def populate(self, zone, target=False, lenient=False):
        self.log.debug(
            'populate: name=%s, target=%s, lenient=%s',
            zone.name,
            target,
            lenient,
        )

        before = len(zone.records)

        done = False
        page_token = None
        while not done:
            resp = self.cdn_service.List(
                ListResourcesRequest(
                    folder_id=self.folder_id, page_token=page_token
                )
            )

            if resp.next_page_token:
                page_token = resp.next_page_token
            else:
                done = True

            for resource in resp.resources:
                self.process_resource(zone, resource, lenient)

        self.log.info('populate: found %s records', len(zone.records) - before)



from google.protobuf.timestamp_pb2 import Timestamp
from yandex.cloud.dns.v1.dns_zone_pb2 import (
    DnsZone,
    PrivateVisibility,
    PublicVisibility,
)

from tests.fixtures import (
    STUB_FOLDER_ID,
    STUB_IDNA_ZONE_NAME,
    STUB_NETWORK_ID,
    STUB_ZONE_NAME,
)

STUB_ZONE_1 = DnsZone(
    id='dnsisei4peetahsak6oh',
    folder_id=STUB_FOLDER_ID,
    created_at=Timestamp(seconds=1710685039, nanos=779000000),
    name="test2",
    zone=STUB_ZONE_NAME,
    private_visibility=PrivateVisibility(network_ids=[STUB_NETWORK_ID]),
)

STUB_ZONE_2 = DnsZone(
    id='dnsh5athi0ha3di4eiv6',
    folder_id=STUB_FOLDER_ID,
    created_at=Timestamp(seconds=1710685015, nanos=601000000),
    name="test",
    zone=STUB_ZONE_NAME,
    private_visibility=PrivateVisibility(),
)

STUB_ZONE_PUBLIC = DnsZone(
    id='dnsg8coh3cheiBahpohk',
    folder_id=STUB_FOLDER_ID,
    created_at=Timestamp(seconds=1710325719, nanos=138000000),
    name="example-com",
    zone=STUB_ZONE_NAME,
    public_visibility=PublicVisibility(),
)


STUB_IDNA_ZONE = DnsZone(
    id='dnshahmeep0hei5aighe',
    folder_id=STUB_FOLDER_ID,
    created_at=Timestamp(seconds=1710779521, nanos=553000000),
    name="idna-test",
    description="An idna test",
    zone=STUB_IDNA_ZONE_NAME,
    private_visibility=PrivateVisibility(),
)



from logging import getLogger

import yandexcloud
from yandex.cloud.certificatemanager.v1.certificate_pb2 import (
    CertificateType,
    ChallengeType,
)
from yandex.cloud.certificatemanager.v1.certificate_service_pb2 import (
    ListCertificatesRequest,
)
from yandex.cloud.certificatemanager.v1.certificate_service_pb2_grpc import (
    CertificateServiceStub,
)

from octodns.record import Record
from octodns.source.base import BaseSource

from octodns_yandex.auth import _AuthMixin
from octodns_yandex.exception import YandexCloudConfigException
from octodns_yandex.version import get_user_agent


class YandexCloudCMSource(_AuthMixin, BaseSource):
    SUPPORTS_GEO = False
    SUPPORTS = {'CNAME', 'TXT'}

    def __init__(
        self,
        id,
        folder_id: str,
        auth_type: str,
        record_type='CNAME',
        record_ttl=3600,
        oauth_token=None,
        iam_token=None,
        sa_key_file=None,
        sa_key=None,
        *args,
        **kwargs,
    ):
        self.log = getLogger(f"YandexCloudCMSource[{id}]")

        self.folder_id = folder_id
        self.record_type = record_type
        if record_type not in self.SUPPORTS:
            raise YandexCloudConfigException('Not supported record_type')
        self.record_ttl = record_ttl

        self.auth_kwargs = self.get_auth_kwargs(
            auth_type, oauth_token, iam_token, sa_key_file, sa_key
        )
        self.log.debug(
            '__init__: folder_id=%s auth_type=%s auth_kwargs=%s',
            self.folder_id,
            auth_type,
            self.auth_kwargs,
        )

        super().__init__(id, *args, **kwargs)

        self.sdk = yandexcloud.SDK(
            user_agent=get_user_agent(), **self.auth_kwargs
        )
        self.cm_service = self.sdk.client(CertificateServiceStub)

    def process_certificate(self, zone, cert, lenient=False):
        if cert.type != CertificateType.MANAGED:
            return

        if not next(
            (x for x in cert.domains if zone.owns(self.record_type, x)), None
        ):
            return

        new_records = {}
        for challenge in cert.challenges:
            if challenge.type != ChallengeType.DNS:
                continue

            challenge_record = challenge.dns_challenge
            if challenge_record.type != self.record_type:
                continue

            if not zone.owns(self.record_type, challenge_record.name):
                continue

            hostname = zone.hostname_from_fqdn(challenge_record.name)
            previous_values = []

            # If there is already same record exists - skip
            # (possible when requesting cert for domain and its wildcard)
            if new_records.get(hostname, None):
                if challenge_record.type == 'CNAME':
                    continue

                prev_data = new_records[hostname].data
                previous_values = prev_data.get('values', [prev_data['value']])

            data = {'type': challenge_record.type, 'ttl': self.record_ttl}
            if challenge_record.type == 'CNAME':
                data['value'] = challenge_record.value
            else:
                data['values'] = [*previous_values, challenge_record.value]

            new_records[hostname] = Record.new(
                zone, hostname, data=data, source=self, lenient=lenient
            )

        for k, record in new_records.items():
            zone.add_record(record)

    def populate(self, zone, target=False, lenient=False):
        self.log.debug(
            'populate: name=%s, target=%s, lenient=%s',
            zone.name,
            target,
            lenient,
        )

        before = len(zone.records)

        done = False
        page_token = None
        while not done:
            resp = self.cm_service.List(
                ListCertificatesRequest(
                    folder_id=self.folder_id, view='FULL', page_token=page_token
                )
            )

            if resp.next_page_token:
                page_token = resp.next_page_token
            else:
                done = True

            for cert in resp.certificates:
                self.process_certificate(zone, cert, lenient)

        self.log.info('populate: found %s records', len(zone.records) - before)



from importlib.metadata import PackageNotFoundError, version

from octodns import __VERSION__ as octodns_version

# TODO: remove __VERSION__ with the next major version.py release
__version__ = __VERSION__ = '0.0.3'

try:
    yandexcloud_version = version("yandexcloud")
except PackageNotFoundError:
    yandexcloud_version = "0.0.0"


def get_base_user_agent():
    return f"octodns/{octodns_version} octodns-yandex/{__version__}"


def get_user_agent():
    return (
        f"{get_base_user_agent()} yandex-cloud-python-sdk/{yandexcloud_version}"
    )



from yandex.cloud.dns.v1.dns_zone_pb2 import RecordSet
from yandex.cloud.dns.v1.dns_zone_service_pb2 import (
    ListDnsZoneRecordSetsResponse,
)

from octodns.record import Ipv6Value

# Value which key equals DNS record type should not be lenient
STUB_RECORDS = {
    'A': RecordSet(
        name="a.example.com.",
        type="A",
        ttl=300,
        data=["127.0.0.1", "127.0.0.2"],
    ),
    'AAAA': RecordSet(
        name="aaaa.example.com.",
        type="AAAA",
        ttl=300,
        # octodns simplifies IPv6
        data=[str(Ipv6Value("2001:0db8:11a3:09d7:1f34:8a2e:07a0:765d"))],
    ),
    'CAA': RecordSet(
        name="caa.example.com.",
        type="CAA",
        ttl=600,
        data=[
            "0 issue ca.example.net",
            "0 issue ca2.example.net",
            "3 asd asd",  # valid in YC
        ],
    ),
    'CAA_quoted': RecordSet(
        name="caa2.example.com.",
        type="CAA",
        ttl=600,
        data=["0 issue \"ca.example.net\""],
    ),
    'CNAME': RecordSet(
        name="cname1.example.com.",
        type="CNAME",
        ttl=3600,
        data=["example.com."],
    ),
    'CNAME_nodot': RecordSet(
        name="cname2.example.com.",
        type="CNAME",
        ttl=3600,
        data=["example.com"],  # valid in YC
    ),
    'CNAME_dot': RecordSet(
        name="cname3.example.com.",
        type="CNAME",
        ttl=600,
        data=["."],  # valid in YC
    ),
    # fmt: off
    'MX': RecordSet(
        name="mx.example.com.",
        type="MX",
        ttl=600,
        data=["10 mx.example.com."],
    ),
    'MX_nodot': RecordSet(
        name="mx2.example.com.",
        type="MX",
        ttl=600,
        data=["10 mx.example.com"],
    ),
    # fmt: on
    'NS': RecordSet(
        name="ns.example.com.",
        type="NS",
        ttl=600,
        data=["ns1.yandexcloud.net.", "ns2.yandexcloud.net."],
    ),
    'NS_root': RecordSet(
        name="example.com.",
        type="NS",
        ttl=3600,
        data=["ns1.yandexcloud.net.", "ns2.yandexcloud.net."],
    ),
    'SOA': RecordSet(
        name="example.com.",
        type="SOA",
        ttl=3600,
        data=[
            "ns1.yandexcloud.net. mx.cloud.yandex.net. 1 10800 900 604800 900"
        ],
    ),
    'PTR': RecordSet(
        name="ptr.example.com.",
        type="PTR",
        ttl=600,
        data=["example.com.", "example2.com."],
    ),
    'PTR_nodot': RecordSet(
        name="ptr2.example.com.",
        type="PTR",
        ttl=600,
        data=["example.com"],  # valid in YC
    ),
    'SRV': RecordSet(
        name="_service._tcp.example.com.",
        type="SRV",
        ttl=600,
        data=["10 70 8080 backend.example.com."],
    ),
    'SRV_nodot': RecordSet(
        name="_service2._tcp.example.com.",
        type="SRV",
        ttl=600,
        data=["10 70 8080 backend.example.com"],
    ),
    'TXT': RecordSet(
        name="txt.example.com.",
        type="TXT",
        ttl=600,
        data=[
            "test3",
            "v=DKIM1\\; k=rsa\\; s=email\\\\; p=MIGfMA0GCSqGSIb3DQEBAQUAA4GNADCBiQKBgQC96HHGGEenw"
            "TqEDp/ZA1ZKDkb89AVyWwyBN2ALLqtnCVffqMm9yd3N6pkW0378Jn1RxYpKttMGa7fwthMwnlN6cbo7h9wXm"
            "SRE/B7kmXtoA9QBX/5rIem11pqiP51YVNJqysMXlE6k/SPvr6KXc49s2X5jIqAE6YjI0jHUe6XDMQIDAQAB",
        ],
    ),
    # fmt: off
    'TXT_quoted': RecordSet(
        name="txt2.example.com.",
        type="TXT",
        ttl=600,
        data=["\"test\""],
    ),
    'ANAME': RecordSet(
        name="aname.example.com.",
        type="ANAME",
        ttl=600,
        data=["example.com."],
    ),
    'ANAME_nodot': RecordSet(
        name="aname2.example.com.",
        type="ANAME",
        ttl=600,
        data=["example.com"],
    ),
    'ANAME_mx': RecordSet(
        name="aname.example.com.",
        type="MX",
        ttl=600,
        data=["20 mx.example.com."],
    ),
    # fmt: on
}

IDNA_LIST_RECORD_SETS_RESPONSE_PAGE_1 = ListDnsZoneRecordSetsResponse(
    next_page_token='',
    record_sets=[
        RecordSet(
            name="xn--b1agh1afp.xn--e1aybc.xn--p1ai.",
            type="A",
            ttl=600,
            data="127.0.0.2",
        ),
        # fmt: off
        RecordSet(
            name="xn--e1aybc.xn--p1ai.",
            type="A",
            ttl=600,
            data="127.0.0.1",
        ),
        # fmt: on
        RecordSet(
            name="xn--e1aybc.xn--p1ai.",
            type="NS",
            ttl=3600,
            data="ns.internal.",
        ),
        RecordSet(
            name="xn--e1aybc.xn--p1ai.",
            type="SOA",
            ttl=3600,
            data="ns.internal. mx.cloud.yandex.net. 1 10800 900 604800 900",
        ),
    ],
)



import json
import subprocess

from octodns_yandex.exception import YandexCloudConfigException

AUTH_TYPE_OAUTH = 'oauth'
AUTH_TYPE_IAM = 'iam'
AUTH_TYPE_METADATA = 'metadata'
AUTH_TYPE_SA_KEY = 'sa-key'
AUTH_TYPE_YC_CLI = 'yc-cli'


class _AuthMixin(object):
    @staticmethod
    def get_auth_kwargs(auth_type, oauth_token, iam_token, sa_key_file, sa_key):
        if auth_type == AUTH_TYPE_OAUTH:
            return {'token': oauth_token}
        elif auth_type == AUTH_TYPE_IAM:
            return {'iam_token': iam_token}
        elif auth_type == AUTH_TYPE_METADATA:
            return {}  # Autoconfigured by SDK
        elif auth_type == AUTH_TYPE_SA_KEY:
            if sa_key_file:
                try:
                    with open(sa_key_file) as infile:
                        sa_key = json.load(infile)
                except OSError as e:
                    raise YandexCloudConfigException(
                        f"Failed to open 'sa_key_file': {sa_key_file}"
                    ) from e
            if (
                sa_key is None
                or 'id' not in sa_key
                or 'service_account_id' not in sa_key
                or 'private_key' not in sa_key
            ):
                raise YandexCloudConfigException(
                    "Provider option 'sa_key' should be dict with fields: id, service_account_id, private_key"
                )
            return {'service_account_key': sa_key}
        elif auth_type == AUTH_TYPE_YC_CLI:
            process = None
            try:
                process = subprocess.run(
                    ['yc', 'config', 'get', 'token'], stdout=subprocess.PIPE
                )
                process.check_returncode()
            except FileNotFoundError as e:
                raise YandexCloudConfigException(
                    'yc binary not found in PATH'
                ) from e
            except subprocess.CalledProcessError as e:
                raise YandexCloudConfigException(
                    f"Failed to get token from yc, exit code: {process.returncode}"
                ) from e

            return {'token': process.stdout.decode('utf-8').strip()}
        else:
            raise YandexCloudConfigException("Unknown auth type")
